@@ -17,6 +17,7 @@ import pickle
 import tempfile
 from functools import wraps
 from datetime import date
+import sys
 
 from flask import request, session, redirect, current_app, url_for
 from werkzeug import url_quote
@@ -24,7 +25,8 @@ from werkzeug import url_quote
 from openid.store.filestore import FileOpenIDStore
 from openid.extensions import ax
 from openid.extensions.sreg import SRegRequest, SRegResponse
-from openid.consumer.consumer import Consumer, SUCCESS, CANCEL
+from openid.consumer.consumer import Consumer, SUCCESS, CANCEL, FAILURE, \
+    SETUP_NEEDED
 from openid.consumer import discover
 
 # python-openid is a really stupid library in that regard, we have
@@ -101,6 +103,12 @@ def softint(x):
     except (ValueError, TypeError):
         return None
 
+def isstring(x):
+    if sys.version_info[0] >= 3:
+        return isinstance(x, str)
+    else:
+        return isinstance(x, basestring)
+
 
 class SessionWrapper(object):
     name_mapping = {
@@ -114,7 +122,10 @@ class SessionWrapper(object):
     def __getitem__(self, name):
         rv = session[self.name_mapping.get(name, name)]
         if isinstance(rv, dict) and len(rv) == 1 and ' p' in rv:
-            return pickle.loads(rv[' p'].encode('utf-8'))
+            try:
+                return pickle.loads(rv[' p'].encode('utf-8'))
+            except:
+                return pickle.loads(rv[' p'])
         return rv
 
     def __setitem__(self, name, value):
@@ -329,7 +340,7 @@ class OpenID(object):
     def __init__(self, app=None, fs_store_path=None, store_factory=None,
                  fallback_endpoint=None, extension_responses=None, safe_roots=None):
         # backwards compatibility support
-        if isinstance(app, basestring):
+        if isstring(app):
             from warnings import warn
             warn(DeprecationWarning('OpenID constructor expects application '
                                     'as first argument now.  If you want to '
@@ -353,7 +364,7 @@ class OpenID(object):
         if not extension_responses:
             extension_responses = []
         self.extension_responses = extension_responses
-        if isinstance(safe_roots, basestring):
+        if isstring(safe_roots):
             self.safe_roots = [safe_roots]
         else:
             self.safe_roots = safe_roots
@@ -487,14 +498,22 @@ class OpenID(object):
                     openid_response, self.extension_responses))
             elif openid_response.status == CANCEL:
                 self.signal_error(u'The request was cancelled')
-                return redirect(self.get_current_url())
-            self.signal_error(u'OpenID authentication error')
+            elif openid_response.status == FAILURE:
+                self.signal_error(u'OpenID authentication failure. Mesage: %s'
+                    % openid_response.message)
+            elif openid_response.status == SETUP_NEEDED:
+                # We should never get here since we do checkid_setup always
+                self.signal_error(u'OpenID setup was needed??')
+            else:
+                # We should also never get here, as this should be exhaustive
+                self.signal_error(u'OpenID authentication weird state: %s' %
+                    openid_response.status)
             return redirect(self.get_current_url())
         return decorated
 
     def try_login(self, identity_url, ask_for=None, ask_for_optional=None, extensions=None):
         """This tries to login with the given identity URL.  This function
-        must be called from the login_handler.  The `ask_for` and 
+        must be called from the login_handler.  The `ask_for` and
         `ask_for_optional`parameter can be a set of values to be asked
         from the openid provider, where keys in `ask_for` are marked as
         required, and keys in `ask_for_optional` are marked as optional.
